import contextlib
import sys

from boa.contracts.base_evm_contract import BoaError
from boa.contracts.vyper.vyper_contract import check_boa_error_matches
from boa.debugger import BoaDebug
from boa.environment import Env
from boa.interpret import (
    from_etherscan,
    load,
    load_abi,
    load_partial,
    loads,
    loads_abi,
    loads_partial,
)
from boa.network import NetworkEnv
from boa.precompile import precompile
from boa.test.strategies import fuzz
from boa.vm.py_evm import enable_pyevm_verbose_logging, patch_opcode

# turn off tracebacks if we are in repl
# https://stackoverflow.com/a/64523765
if hasattr(sys, "ps1"):  # pragma: no cover
    pass
    # sys.tracebacklimit = 0

env = Env.get_singleton()


@contextlib.contextmanager
def swap_env(new_env):
    old_env = env
    try:
        set_env(new_env)
        yield
    finally:
        set_env(old_env)


def set_env(new_env):
    global env
    env = new_env

    Env._singleton = new_env


def set_browser_env(address=None):
    """Set the environment to use the browser's network in Jupyter/Colab"""
    # import locally because jupyter is generally not installed
    from boa.integrations.jupyter import BrowserEnv

    set_env(BrowserEnv(address))


def set_network_env(url):
    """Set the environment to use a custom network URL"""
    set_env(NetworkEnv.from_url(url))


def reset_env():
    set_env(Env())


def _breakpoint(computation):
    BoaDebug(computation).start()


patch_opcode(0xA6, _breakpoint)


@contextlib.contextmanager
def reverts(*args, **kwargs):
    try:
        yield
        raise ValueError("Did not revert")
    except BoaError as b:
        if args or kwargs:
            check_boa_error_matches(b, *args, **kwargs)


def eval(code):
    return loads("").eval(code)


def _jupyter_server_extension_points() -> list[dict]:
    """
    Returns a list of dictionaries with metadata describing
    where to find the `_load_jupyter_server_extension` function.
    """
    return [{"module": "boa.integrations.jupyter"}]



import contextlib
import inspect
import textwrap
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import PurePath
from typing import Any, Optional

import vyper.ir.optimizer
from eth.exceptions import Halt
from eth.exceptions import Revert as VMRevert
from eth.exceptions import WriteProtection
from eth_hash.auto import keccak
from vyper.ast.nodes import VyperNode
from vyper.codegen.ir_node import IRnode
from vyper.compiler.phases import CompilerData
from vyper.evm.opcodes import OPCODES
from vyper.ir.compile_ir import getpos
from vyper.utils import unsigned_to_signed

from boa.util.lrudict import lrudict
from boa.vm.fast_mem import FastMem
from boa.vm.utils import ceil32, to_bytes, to_int

_keccak_cache = lrudict(256)


# note: This is used in the generated code for `Sha3_64` below.
def keccak256(x):
    return _keccak_cache.setdefault_lambda(x, keccak)


def _mkalphanum(string):
    # map a string to only-alphanumeric chars
    return "".join([c if c.isalnum() else "_" for c in string])


@dataclass
class _Line:
    indentation_level: int
    line: str

    def show(self, indenter=" "):
        return indenter * self.indentation_level + self.line


@dataclass
class PythonBuilder:
    cur_indentation_level: int = 0
    lines: list[_Line] = field(default_factory=list)

    def extend(self, source_code):
        source_code = textwrap.dedent(source_code)
        for line in source_code.splitlines():
            self.append(line)

    def append(self, source_code):
        self.lines.append(_Line(self.cur_indentation_level, source_code))

    def get_output(self):
        return "\n".join(line.show() for line in self.lines)

    @contextlib.contextmanager
    def block(self, entry):
        self.append(entry + ":")
        self.cur_indentation_level += 4
        yield
        self.cur_indentation_level -= 4


@dataclass
class FrameInfo:
    current_slot: int = 0  # basically the de bruijn index
    slots: dict[str, int] = field(default_factory=lambda: {})
    types: dict[str, type] = field(default_factory=lambda: {})


_global_id = 0


@dataclass
class CompileContext:
    # include CompilerData - we need this to get immutable section size
    vyper_compiler_data: CompilerData
    uuid: str = field(init=False)
    labels: dict[str, "IRExecutor"] = field(default_factory=dict)
    unique_symbols: set[str] = field(default_factory=set)
    frames: list[FrameInfo] = field(default_factory=lambda: [FrameInfo()])
    builder: PythonBuilder = field(default_factory=PythonBuilder)
    var_id: int = -1

    def __post_init__(self):
        # use a global bc the generated functions need to be unique
        global _global_id
        self.uuid = str(_global_id)
        _global_id += 1

    @property
    def local_vars(self):
        return self.frames[-1].slots

    def freshvar(self, name=""):
        self.var_id += 1
        return f"var_{name}_{self.var_id}"

    @cached_property
    def contract_name(self):
        return _mkalphanum(PurePath(self.vyper_compiler_data.contract_path).name)

    def translate_label(self, label):
        return _mkalphanum(f"{self.contract_name}_{self.uuid}_{label}")

    def add_unique_symbol(self, symbol):
        if symbol in self.unique_symbols:  # pragma: no cover
            raise ValueError(
                "duplicated symbol {symbol}, this is likely a bug in vyper!"
            )
        self.unique_symbols.add(symbol)

    def add_label(self, labelname, executor):
        if labelname in self.labels:  # pragma: no cover
            raise ValueError("duplicated label: {labelname}")
        self.labels[labelname] = executor

    @contextlib.contextmanager
    def allocate_local_frame(self):
        frame = FrameInfo()
        self.frames.append(frame)
        yield  # frame
        self.frames.pop()

    @contextlib.contextmanager
    def variables(self, vars_list):
        # allocate variables in vars_list, assigning them each
        # a new slot.
        shadowed = {}
        frame = self.frames[-1]
        for varname in vars_list:
            shadowed[varname] = frame.slots.get(varname)
            frame.slots[varname] = frame.current_slot
            frame.current_slot += 1

        yield

        for varname in vars_list:
            frame.current_slot -= 1
            if shadowed[varname] is None:
                del frame.slots[varname]
            else:
                frame.slots[varname] = shadowed[varname]


StackItem = int | bytes


mapper = {int: "to_int", bytes: "to_bytes", StackItem: ""}


class ExecutionContext:
    __slots__ = "computation"

    def __init__(self, computation):
        self.computation = computation


class IRExecutor:
    __slots__ = ("args", "compile_ctx", "_exec", "ir_node")

    # the type produced when executing this node
    _type: Optional[type] = None  # | int | bytes
    _is_static = True  # can be used from a STATICCALL context

    def __init__(self, ir_node, compile_ctx, *args):
        self.ir_node = ir_node
        self.args = args
        self.compile_ctx = compile_ctx

    @cached_property
    def name(self):
        return self._name

    def _compile_args(self, argnames):
        assert len(self.args) == len(argnames) == len(self._sig), (
            type(self),
            self.args,
            argnames,
            self._sig,
            self.ir_node,
        )
        for out, arg, typ in reversed(list(zip(argnames, self.args, self._sig))):
            arg.compile(out=out, out_typ=typ)

    @cached_property
    def builder(self):
        return self.compile_ctx.builder

    def analyze(self):
        self.args = [arg.analyze() for arg in self.args]
        return self

    # compile an IR "expr". called for its side effects on builder
    def compile(self, out=None, out_typ=None):
        # do a bit of metaprogramming to infer how to compile the args
        if hasattr(self, "_argnames"):
            argnames = self._argnames
        else:
            argnames = inspect.getargs(self._compile.__code__).args
            assert argnames[0] == "self"
            argnames = argnames[1:]

        argnames = [self.compile_ctx.freshvar(x) for x in argnames]

        self._compile_args(argnames)

        if hasattr(self, "_name"):
            self.builder.append(f"# {self.name}")

        if not self._is_static:
            _ = WriteProtection  # make flake8 happy
            self.builder.extend(
                """
            if VM.msg.is_static:
                raise WriteProtection(
                    "Cannot modify state while inside of a STATICCALL context"
                )
            """
            )

        res = self._compile(*argnames)

        if res is None:
            assert out is None, (type(self), self, out, argnames)
            return

        if out is not None:
            # squash F401 lint complaint about import
            _ = to_int, to_bytes
            if self._type != out_typ:
                res = f"{mapper[out_typ]}({res})"
            self.builder.append(f"{out} = {res}")
        else:
            self.builder.append(res)

    def _compile(self, context):  # pragma: no cover
        raise RuntimeError("must be overridden in subclass!")

    def compile_main(self, contract_name=""):
        self.builder.extend("import vyper.utils\nimport _operator")

        main_name = self.compile_ctx.translate_label("main")
        with self.builder.block(f"def {main_name}(CTX)"):
            self.builder.append("VM = CTX.computation")
            self.compile()

        for func in self.compile_ctx.labels.values():
            self.builder.extend("\n\n")
            func.compile_func()

        py_file = f"{contract_name}{self.compile_ctx.uuid}.py"

        # uncomment for debugging the python code:
        # with open(py_file, "w") as f:
        #    print(self.builder.get_output(), file=f)

        py_bytecode = compile(self.builder.get_output(), py_file, "exec")
        exec(py_bytecode, globals())

        self._exec = globals()[main_name]

    def exec(self, computation):
        computation._memory = FastMem()
        execution_ctx = ExecutionContext(computation)
        self._exec(execution_ctx)


@dataclass
class IntExecutor(IRExecutor):
    compile_ctx: CompileContext
    _int_value: int
    _type: type = int

    def __post_init__(self):
        assert -(2**255) <= self._int_value < 2**256
        self._int_value = _wrap256(self._int_value)  # wrap, could be negative
        self.args = self._sig = ()

    def __repr__(self):
        return hex(self._int_value)

    def analyze(self):
        return self

    def _compile(self):
        return repr(self)


@dataclass
class StringExecutor(IRExecutor):
    compile_ctx: CompileContext
    _str_value: str

    @property
    def _type(self):
        raise RuntimeError("should have been analyzed!")

    def __post_init__(self):
        self.args = self._sig = ()

    def __repr__(self):
        return repr(self._str_value)

    def analyze(self):
        slot = self.compile_ctx.local_vars[self._str_value]
        return VariableExecutor(self.compile_ctx, self._str_value, slot)


@dataclass
class VariableExecutor(IRExecutor):
    compile_ctx: CompileContext
    varname: str
    var_slot: int

    # optimization assumption: most variables that
    # will be hotspots need to be ints.
    _type: type = int

    def __post_init__(self):
        self.args = self._sig = ()

    def __repr__(self):
        return f"var({self.varname})"

    @cached_property
    def out_name(self):
        slot = self.var_slot
        ret = f"__user_{self.varname}"
        if slot > 0:
            ret += f"_{slot}"
        return ret

    def analyze(self):
        raise RuntimeError("Should not appear during analysis!")

    def _compile(self):
        return self.out_name


@dataclass
class OpcodeInfo:
    # model of an opcode from vyper.evm.opcodes
    mnemonic: str
    opcode: int  # opcode number ex. 0x01 for ADD
    consumes: int  # number of stack items this consumes
    produces: int  # number of stack items this produces, must be 0 or 1
    _gas_estimate: int  # in vyper.evm.opcodes but probably not useful for us

    def __post_init__(self):
        assert 0 <= self.opcode < 256
        assert self.produces in (0, 1)

    @classmethod
    def from_opcode_info(cls, mnemonic, opcode_info):
        # info from vyper.evm.opcodes
        opcode, consumes, produces, gas_estimate = opcode_info
        return cls(mnemonic, opcode, consumes, produces, gas_estimate)

    @classmethod
    def from_mnemonic(cls, mnemonic):
        mnemonic = mnemonic.upper()
        return cls.from_opcode_info(mnemonic, OPCODES[mnemonic])


# an executor for evm opcodes which dispatches into py-evm
class OpcodeIRExecutor(IRExecutor):
    _type: type = StackItem  # type: ignore

    def __init__(self, name: str, opcode_info: OpcodeInfo, *args):
        self.opcode_info = opcode_info

        # to differentiate from implemented codes
        self._name = "__" + name + "__"

        super().__init__(*args)

    def __repr__(self):
        args = ",".join(repr(arg) for arg in self.args)
        return f"{self.name}({args})"

    @cached_property
    def _sig(self):
        # TODO figure out the type to avoid calling to_int
        return tuple(int for _ in range(self.opcode_info.consumes))

    @cached_property
    def _argnames(self):
        def mkargname(i):
            return f"__{self.opcode_info.mnemonic.lower()}_arg{i}"

        return tuple(mkargname(i) for i in range(self.opcode_info.consumes))

    def _compile(self, *args):
        for arg in reversed(args):
            self.builder.append(f"VM.stack_push_int({arg})")

        opcode = hex(self.opcode_info.opcode)
        self.builder.append(f"VM.opcodes[{opcode}].__call__(CTX.computation)")
        if self.opcode_info.produces:
            return "VM.stack_pop1_any()"


_executors = {}


# decorator to register an executor class in the _executors dict.
def executor(cls):
    _executors[cls._name] = cls
    return cls


def _wrap256(x):
    return x % 2**256


def wrap256_str(x_str):
    return f"(({x_str}) % 2**256)"


def _as_signed(x):
    return unsigned_to_signed(x, 256, strict=True)


# XXX: some of these do not need to be wrapped
class UnsignedBinopExecutor(IRExecutor):
    __slots__ = ("_name", "_op")
    _sig = int, int
    _type: type = int

    @cached_property
    def funcname(self):
        return self._op.__module__ + "." + self._op.__name__

    def _compile(self, x, y):
        return wrap256_str(f"{self.funcname}({x}, {y})")


class SignedBinopExecutor(UnsignedBinopExecutor):
    def _compile(self, x, y):
        return wrap256_str(f"{self.funcname}(_as_signed({x}), _as_signed({y}))")


# for binops, just use routines from vyper optimizer
for opname, (op, _, unsigned) in vyper.ir.optimizer.arith.items():
    base = UnsignedBinopExecutor if unsigned else SignedBinopExecutor
    nickname = opname.capitalize()
    _executors[opname] = type(nickname, (base,), {"_op": op, "_name": opname})


# shift instructions have opposite operand order from operator.*shift functions.
@executor
class Shr(IRExecutor):
    _name = "shr"
    _sig = (int, int)
    _type: type = int

    def _compile(self, bits, val):
        return f"{val} >> {bits}"


@executor
class Sar(IRExecutor):
    _name = "sar"
    _sig = (int, int)
    _type: type = int

    def _compile(self, bits, val):
        # wrap256 to get back into unsigned land
        return wrap256_str(f"_as_signed({val}) >> {bits}")


@executor
class Shl(IRExecutor):
    _name = "shl"
    _sig = (int, int)
    _type: type = int

    def _compile(self, bits, val):
        return wrap256_str(f"{val} << {bits}")


@executor
class Select(IRExecutor):
    _name = "select"
    _sig = (int, StackItem, StackItem)

    @cached_property
    def _type(self):
        _, x, y = self.args
        if x._type == y._type:
            return x._type
        return StackItem

    def _compile(self, test, x, y):
        return f"{x} if {test} else {y}"


_NULL_BYTE = repr(b"\x00")


@executor
class Caller(IRExecutor):
    _name = "caller"
    _sig = ()
    _type: type = bytes

    def _compile(self):
        return "VM.msg.sender"


@executor
class CalldataSize(IRExecutor):
    _name = "calldatasize"
    _sig = ()
    _type: type = int

    def _compile(self):
        return "len(VM.msg.data)"


@executor
class CallValue(IRExecutor):
    _name = "callvalue"
    _sig = ()
    _type: type = int

    def _compile(self):
        return "VM.msg.value"


@executor
class CalldataLoad(IRExecutor):
    _name = "calldataload"
    _sig = (int,)
    _type: type = bytes

    def _compile(self, ptr):
        self.builder.extend(
            f"""
            val = bytes(VM.msg.data[{ptr} : {ptr} + 32])
            """
        )
        return f"val.ljust(32, {_NULL_BYTE})"


@executor
class CalldataCopy(IRExecutor):
    _name = "calldatacopy"
    _sig = (int, int, int)

    def _compile(self, dst, src, size):
        self.builder.extend(
            f"""
            val = bytes(VM.msg.data[{src} : {src} + {size}])
            val = val.ljust({size}, {_NULL_BYTE})

            VM._memory.extend({dst}, {size})
            VM.memory_write({dst}, {size}, val)
            """
        )


@executor
class MLoad(IRExecutor):
    _name = "mload"
    _sig = (int,)
    _type: type = int

    def _compile(self, ptr):
        self.builder.append(f"VM._memory.extend({ptr}, 32)")
        return f"VM._memory.read_word({ptr})"


@executor
class MStore(IRExecutor):
    _name = "mstore"
    _sig = (int, int)

    def _compile(self, ptr, val):
        self.builder.extend(
            f"""
        VM._memory.extend({ptr}, 32)
        VM._memory.write_word({ptr}, {val})
        """
        )


class _CodeLoader(IRExecutor):
    @cached_property
    def immutables_size(self):
        compiler_data = self.compile_ctx.vyper_compiler_data
        return compiler_data.global_ctx.immutable_section_bytes


@executor
class DLoad(_CodeLoader):
    _name = "dload"
    _sig = (int,)
    _type: type = bytes

    def _compile(self, ptr):
        assert self.immutables_size > 0
        self.builder.extend(
            f"""
        code_start_position = {ptr} - {self.immutables_size} + len(VM.code)

        with VM.code.seek(code_start_position):
            ret = VM.code.read(32)
        """
        )
        return f"ret.ljust(32, {_NULL_BYTE})"


@executor
class DLoadBytes(_CodeLoader):
    _name = "dloadbytes"
    _sig = (int, int, int)

    def _compile(self, dst, src, size):
        assert self.immutables_size > 0

        # adapted from py-evm codecopy, but without gas metering, then
        # mess with the start position
        self.builder.extend(
            f"""
        code_start_position = {src} - {self.immutables_size} + len(VM.code)
        VM._memory.extend({dst}, {size})

        with VM.code.seek(code_start_position):
            code_bytes = VM.code.read({size})

        padded_code_bytes = code_bytes.ljust({size}, {_NULL_BYTE})

        VM.memory_write({dst}, {size}, padded_code_bytes)
        """
        )


@executor
class SLoad(IRExecutor):
    _name = "sload"
    _sig = (int,)
    _type = int

    def _compile(self, slot):
        return f"""VM.state.get_storage(address=VM.msg.storage_address, slot={slot})"""


@executor
class SStore(IRExecutor):
    _name = "sstore"
    _is_static = False
    _sig = (int, int)
    _type = int

    def _compile(self, slot, value):
        self.builder.extend(
            f"""
            VM.env._trace_sstore(VM.msg.storage_address, {slot})
            VM.state.set_storage(address=VM.msg.storage_address, slot={slot}, value={value})
            """
        )


@executor
class Sha3_64(IRExecutor):
    _name = "sha3_64"
    _sig = (bytes, bytes)

    # we need to trace for downstream to reverse engineer mappings
    def _compile(self, arg1, arg2):
        self.builder.extend(
            f"""
        preimage = {arg1}.rjust(32, {_NULL_BYTE}) + {arg2}.rjust(32, {_NULL_BYTE})
        image = keccak256(preimage)
        VM.env._trace_sha3_preimage(preimage, image)
        """
        )
        return "image"


@executor
class Sha3_32(IRExecutor):
    _name = "sha3_32"
    _sig = (bytes,)

    def _compile(self, arg):
        self.builder.extend(
            f"""
        preimage = {arg}.rjust(32, {_NULL_BYTE})
        """
        )
        return "keccak256(preimage)"


class _LogN(IRExecutor):
    @cached_property
    def _argnames(self):
        return ("ofst", "size") + tuple(f"log_arg{i}" for i in range(self.N))

    @cached_property
    def _sig(self):
        return (int, int) + tuple(int for _ in range(self.N))

    def _compile(self, ofst, size, *topics):
        # write out tuple strings correctly, always need trailing comma
        topics = [f"{topic}," for topic in topics]
        self.builder.extend(
            f"""
            VM._memory.extend({ofst}, {size})
            log_data = VM.memory_read_bytes({ofst}, {size})
            VM.add_log_entry(
                account=VM.msg.storage_address,
                topics=({" ".join(topics)}),
                data=log_data,
            )
            """
        )


# generate log0..log4
for i in (0, 1, 2, 3, 4):
    opname = f"log{i}"
    _executors[opname] = type(opname.capitalize(), (_LogN,), {"N": i, "_name": opname})


@executor
class Ceil32(IRExecutor):
    _name = "ceil32"
    _sig = (int,)
    _type: type = int

    def _compile(self, x):
        _ = ceil32  # typing hint
        return f"ceil32({x})"


@executor
class IsZero(IRExecutor):
    _name = "iszero"
    _sig = (int,)
    _type: type = int

    def _compile(self, x):
        return f"({x} == 0)"


@executor
class Pass(IRExecutor):
    _name = "pass"
    _sig = ()
    _argnames = ()

    def _compile(self):
        self.builder.append("pass")


@executor
class Seq(IRExecutor):
    _name = "seq"

    def compile(self, out=None, out_typ=None):
        if len(self.args) == 0:
            self.builder.append("pass")
            return
        for i, arg in enumerate(self.args):
            if i + 1 < len(self.args):
                # don't accidentally assign
                arg.compile(out=None)
            else:
                return arg.compile(out=out, out_typ=out_typ)
        else:  # pragma: no cover
            raise RuntimeError("loop should have broken")


@executor
class Repeat(IRExecutor):
    _name = "repeat"

    def compile(self, out=None, out_typ=None):
        i_var, start, rounds, rounds_bound, body = self.args

        startname = self.compile_ctx.freshvar("start")
        roundsname = self.compile_ctx.freshvar("rounds")
        start.compile(startname, out_typ=int)
        rounds.compile(roundsname, out_typ=int)

        end = f"{startname} + {roundsname}"

        self.builder.append(f"assert {roundsname} <= {rounds_bound}")
        with self.builder.block(f"for {i_var.out_name} in range({startname}, {end})"):
            body.compile()

    def analyze(self):
        i_name, start, rounds, rounds_bound, body = self.args

        # analyze start and rounds before shadowing.
        start = start.analyze()
        rounds = rounds.analyze()

        with self.compile_ctx.variables([i_name._str_value]):
            i_var = i_name.analyze()
            body = body.analyze()
        self.args = i_var, start, rounds, rounds_bound, body
        return self


@executor
class If(IRExecutor):
    _name = "if"

    # override `compile()` so we can get the correct lazy behavior
    def compile(self, out=None, out_typ=None):
        orelse = None
        if len(self.args) == 3:
            test, body, orelse = self.args
        else:
            test, body = self.args

        testname = self.compile_ctx.freshvar("test")
        test.compile(testname, out_typ=int)

        with self.builder.block(f"if bool({testname})"):
            body.compile(out, out_typ)

        if orelse:
            with self.builder.block("else"):
                orelse.compile(out, out_typ)


@executor
class Assert(IRExecutor):
    _name = "assert"
    _sig = (int,)

    def _compile(self, test):
        _ = VMRevert  # make flake8 happy
        self.builder.extend(
            f"""
        if not bool({test}):
            VM.vyper_source_pos = {repr(_get_ir_pos(self.ir_node))}
            VM.vyper_error_msg = {repr(self.ir_node.error_msg)}
            raise VMRevert("")  # venom assert
        """
        )


@executor
class _IRRevert(IRExecutor):
    _name = "revert"
    _sig = (int, int)

    def _compile(self, ptr, size):
        self.builder.extend(
            f"""
            VM.output = VM.memory_read_bytes({ptr}, {size})
            VM.vyper_source_pos = {repr(_get_ir_pos(self.ir_node))}
            VM.vyper_error_msg = {repr(self.ir_node.error_msg)}
            raise VMRevert(VM.output)  # venom revert
        """
        )


@executor
class Return(IRExecutor):
    _name = "return"
    _sig = (int, int)

    def _compile(self, ptr, size):
        _ = Halt  # make flake8 happy
        self.builder.extend(
            f"""
            VM.output = VM.memory_read_bytes({ptr}, {size})
            raise Halt("")  # return
        """
        )


@executor
class Stop(IRExecutor):
    _name = "stop"
    _sig = ()

    def _compile(self):
        self.builder.extend(
            """
            raise Halt("")  # return
            """
        )


@executor
class VarList(IRExecutor):
    _name = "var_list"


@executor
class Goto(IRExecutor):
    _name = "goto"

    is_return_stmt = False

    def analyze(self):
        self.label = self.compile_ctx.translate_label(self.args[0]._str_value)

        # just get the parameters, leaving the label in self.args
        # messes with downstream machinery which tries to analyze the label.
        runtime_args = []
        for arg in self.args[1:]:
            if isinstance(arg, Symbol):
                # we don't need to push the return pc on the way in.
                continue

            runtime_args.append(arg.analyze())

        self.args = runtime_args

        return self

    @cached_property
    def _argnames(self):
        if self.is_return_stmt:
            return ()
        return self.compile_ctx.labels[self.label].analyzed_param_names

    @cached_property
    def _type(self):
        if self.is_return_stmt:
            return None
        return self.compile_ctx.labels[self.label]._type

    @cached_property
    def _sig(self):
        # optimization assumption: they all need to be ints
        return tuple(int for _ in self._argnames)

    def _compile(self, *args):
        argnames = self._argnames
        assert len(argnames) == len(self.args)

        args_str = ", ".join(["CTX"] + list(args))
        return f"{self.label}({args_str})"


@executor
class ExitTo(Goto):
    # exit_to is similar but it is known to end execution of this subroutine
    _name = "exit_to"

    def analyze(self):
        # small helper function
        def _is_return_pc(arg):
            return isinstance(arg, StringExecutor) and arg._str_value == "return_pc"

        if _is_return_pc(self.args[0]):
            assert len(self.args) == 1
            self.is_return_stmt = True

        # strip out return_pc args, we don't need actually need to
        # generate any code for it
        self.args = [arg for arg in self.args if not _is_return_pc(arg)]

        if len(self.args) == 0:
            # it's not really a goto, it's a return statement
            # skip super.analyze() as it will choke on no args
            assert self.is_return_stmt
            return self

        return super().analyze()

    def _compile(self, *args):
        if self.is_return_stmt:
            # straight return
            # skip super._compile() as it will choke on no args
            assert len(self.args) == 0
            self.builder.append("return")
            return

        # execute the subroutine and then return
        # (probably not necessary to return the result but may as well)
        subroutine_call = super()._compile(*args)
        return f"return {subroutine_call}"


@executor
class CleanupRepeat(IRExecutor):
    # a no-op from our perspective
    _name = "cleanup_repeat"
    _sig = ()
    _argnames = ()

    def _compile(self):
        self.builder.append("pass")


@executor
class Break(IRExecutor):
    _name = "break"
    _sig = ()
    _argnames = ()

    def _compile(self):
        self.builder.append("break")


@executor
class Continue(IRExecutor):
    _name = "continue"
    _sig = ()
    _argnames = ()

    def _compile(self):
        self.builder.append("continue")


@executor
class Label(IRExecutor):
    _name = "label"

    @cached_property
    def analyzed_param_names(self):
        _, var_list, _ = self.args
        return [x.out_name for x in var_list.args if x.varname != "return_pc"]

    def analyze(self):
        name, var_list, body = self.args

        # use translate_label to ensure no collisions across compilations
        name._str_value = self.compile_ctx.translate_label(name._str_value)

        self.labelname = name._str_value

        self.compile_ctx.add_label(self.labelname, self)

        with self.compile_ctx.allocate_local_frame():
            params = [param._str_value for param in var_list.args]
            with self.compile_ctx.variables(params):
                var_list = var_list.analyze()
                body = body.analyze()

        self.args = name, var_list, body

        self._type = body._type

        return self

    def compile(self, **kwargs):
        pass

    def compile_func(self):
        _, _, body = self.args
        params_str = ", ".join(["CTX"] + self.analyzed_param_names)
        with self.builder.block(f"def {self.labelname}({params_str})"):
            self.builder.append("VM = CTX.computation")
            body.compile()


@executor
class Symbol(IRExecutor):
    # in IR, a "symbol" is a label which needs to be pushed to the
    # stack for the calling convention.
    _name = "symbol"


@executor
class UniqueSymbol(IRExecutor):
    _name = "unique_symbol"

    def analyze(self):
        # we don't really need to do this analysis since vyper should
        # have done it already, but doesn't hurt to be a little paranoid
        symbol = self.args[0]._str_value
        self.compile_ctx.add_unique_symbol(symbol)
        return self

    def compile(self, **kwargs):
        pass


@executor
class With(IRExecutor):
    _name = "with"

    # variable names can be shadowed, so we need to do a bit of
    # analysis to find unshadowed names
    def analyze(self):
        varname = self.args[0]._str_value
        val = self.args[1].analyze()  # analyze before shadowing

        with self.compile_ctx.variables([varname]):
            variable = self.args[0].analyze()
            body = self.args[2].analyze()

            self.args = (variable, val, body)

        self._type = body._type

        return self

    def compile(self, out=None, out_typ=None):
        variable, val, body = self.args
        # optimization assumption: most variables that
        # will be hotspots need to be ints.
        val.compile(out=variable.out_name, out_typ=int)
        return body.compile(out=out, out_typ=out_typ)


@executor
class Set(IRExecutor):
    _name = "set"

    def compile(self, **kwargs):
        variable, val = self.args
        val.compile(out=variable.out_name, out_typ=int)


def _ensure_ast_source(
    ir_node: IRnode, ast_source: VyperNode = None, error_msg: str = None
):
    if ir_node.ast_source is None:
        ir_node.ast_source = ast_source
    if ir_node.error_msg is None:
        ir_node.error_msg = error_msg
    for arg in ir_node.args:
        _ensure_ast_source(arg, ir_node.ast_source, ir_node.error_msg)


def executor_from_ir(ir_node, vyper_compiler_data) -> Any:
    _ensure_ast_source(ir_node)
    ctx = CompileContext(vyper_compiler_data)
    ret = _executor_from_ir(ir_node, ctx)

    ret = ret.analyze()

    # TODO: rename this, this is "something.vy", but we maybe want
    # "something.py <compiled from .vy>"
    ret.compile_main(ctx.contract_name)
    return ret


def _executor_from_ir(ir_node, compile_ctx) -> Any:
    instr = ir_node.value
    if isinstance(instr, int):
        return IntExecutor(compile_ctx, instr)

    args = [_executor_from_ir(arg, compile_ctx) for arg in ir_node.args]

    if instr in _executors:
        return _executors[instr](ir_node, compile_ctx, *args)

    if (mnemonic := instr.upper()) in OPCODES:
        opcode_info = OpcodeInfo.from_mnemonic(mnemonic)
        return OpcodeIRExecutor(instr, opcode_info, ir_node, compile_ctx, *args)

    assert len(ir_node.args) == 0, ir_node
    assert isinstance(ir_node.value, str)
    return StringExecutor(compile_ctx, ir_node.value)


def _get_ir_pos(ir_node):
    if ir_node.ast_source is None:
        return None
    return getpos(ir_node.ast_source)



from functools import cached_property

import coverage.plugin
import vyper.ast as vy_ast
from vyper.ir import compile_ir

import boa.interpret
from boa.contracts.vyper.ast_utils import get_fn_ancestor_from_node
from boa.environment import Env


def coverage_init(registry, options):
    plugin = TitanoboaPlugin(options)
    registry.add_file_tracer(plugin)
    registry.add_configurer(plugin)

    # set on the class so that reset_env() doesn't disable tracing
    Env._coverage_enabled = True


class TitanoboaPlugin(coverage.plugin.CoveragePlugin):
    def __init__(self, options):
        pass

    def configure(self, config):
        config.get_option("run:source_pkgs").append("boa.environment")

    def file_tracer(self, filename):
        if filename.endswith("boa/environment.py"):
            return TitanoboaTracer()

    def file_reporter(self, filename):
        if filename.endswith(".vy"):
            return TitanoboaReporter(filename)


class TitanoboaTracer(coverage.plugin.FileTracer):
    def __init__(self):
        pass

    # coverage.py requires us to inspect the python call frame to
    # see what line number to produce. we hook into specially crafted
    # Env._hook_trace_pc which is called for every pc if coverage is
    # enabled, and then back out the contract and lineno information
    # from there.

    def _valid_frame(self, frame):
        if hasattr(frame.f_code, "co_qualname"):
            # Python>=3.11
            code_qualname = frame.f_code.co_qualname
            return code_qualname == Env._hook_trace_computation.__qualname__

        else:
            # in Python<3.11 we don't have co_qualname, so try hard to
            # find a match anyways. (this might fail if for some reason
            # the executing env has a monkey-patched _hook_trace_computation
            # or something)
            env = Env.get_singleton()
            return frame.f_code == env._hook_trace_computation.__code__

    def _contract_for_frame(self, frame):
        if not self._valid_frame(frame):
            return None
        return frame.f_locals["contract"]

    def dynamic_source_filename(self, filename, frame):
        contract = self._contract_for_frame(frame)
        if contract is None or contract.filename is None:
            return None

        return str(contract.filename)

    def has_dynamic_source_filename(self):
        return True

    # https://coverage.rtfd.io/en/stable/api_plugin.html#coverage.FileTracer.line_number_range
    def line_number_range(self, frame):
        contract = self._contract_for_frame(frame)
        if contract is None:
            return (-1, -1)

        if (pc := frame.f_locals.get("_pc")) is None:
            return (-1, -1)

        pc_map = contract.source_map["pc_raw_ast_map"]

        node = pc_map.get(pc)
        if node is None:
            return (-1, -1)

        start_lineno = node.lineno
        # end_lineno = node.end_lineno
        # note: `return start_lineno, end_lineno` doesn't seem to work.
        return start_lineno, start_lineno

    # XXX: dynamic context. return function name or something
    def dynamic_context(self, frame):
        pass


# helper function. null returns get optimized directly into a jump
# to function cleanup which maps to the parnet FunctionDef ast.
def _is_null_return(ast_node):
    match ast_node:
        case vy_ast.Return(value=None):
            return True
    return False


class TitanoboaReporter(coverage.plugin.FileReporter):
    def __init__(self, filename, env=None):
        super().__init__(filename)

    @cached_property
    def _compiler_data(self):
        return boa.interpret.compiler_data(self.source(), self.filename)

    def arcs(self):
        ret = set()

        for ast_node in self._compiler_data.vyper_module.get_descendants():
            if isinstance(ast_node, vy_ast.If):
                fn_node = get_fn_ancestor_from_node(ast_node)

                # one arc is directly into the body
                arc_true = ast_node.body[0].lineno
                if _is_null_return(ast_node.body[0]):
                    arc_true = fn_node.lineno
                ret.add((ast_node.lineno, arc_true))

                # the other arc is to the end of the if statement
                # try hard to find the next executable line.
                children = ast_node._parent.get_children()
                for node, next_ in zip(children, children[1:]):
                    if id(node) == id(ast_node):
                        arc_false = next_.lineno
                        break
                else:
                    # the if stmt was the last stmt in the enclosing scope.
                    arc_false = ast_node._parent.end_lineno + 1

                # unless there is an else or elif. then the other
                # arc is to the else/elif statement.
                if ast_node.orelse:
                    arc_false = ast_node.orelse[0].lineno

                # return cases:
                # if it's past the end of the fn it's an implicit return
                if arc_false > fn_node.end_lineno:
                    arc_false = fn_node.lineno
                # or it's an explicit return
                if ast_node.orelse and _is_null_return(ast_node.orelse[0]):
                    arc_false = fn_node.lineno

                ret.add((ast_node.lineno, arc_false))

        return ret

    def exit_counts(self):
        ret = {}
        for ast_node in self._compiler_data.vyper_module.get_descendants(vy_ast.If):
            ret[ast_node.lineno] = 2
        return ret

    @cached_property
    def _lines(self):
        ret = set()
        c = self._compiler_data

        # source_map should really be in CompilerData
        _, source_map = compile_ir.assembly_to_evm(c.assembly_runtime)

        for node in source_map["pc_raw_ast_map"].values():
            ret.add(node.lineno)

        return ret

    # OVERRIDES
    def lines(self):
        return self._lines



import json
import sys
import textwrap
from importlib.abc import MetaPathFinder
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

import vyper
from vyper.cli.vyper_compile import get_search_paths
from vyper.compiler.input_bundle import (
    ABIInput,
    CompilerInput,
    FileInput,
    FilesystemInputBundle,
)
from vyper.compiler.phases import CompilerData
from vyper.compiler.settings import Settings, anchor_settings
from vyper.semantics.types.module import ModuleT
from vyper.utils import sha256sum

from boa.contracts.abi.abi_contract import ABIContractFactory
from boa.contracts.vyper.vyper_contract import (
    VyperBlueprint,
    VyperContract,
    VyperDeployer,
)
from boa.environment import Env
from boa.explorer import fetch_abi_from_etherscan
from boa.util.abi import Address
from boa.util.disk_cache import DiskCache

if TYPE_CHECKING:
    from vyper.semantics.analysis.base import ImportInfo

_Contract = Union[VyperContract, VyperBlueprint]


_disk_cache = None
_search_path = None


def set_search_path(path: list[str]):
    global _search_path
    _search_path = path


def set_cache_dir(cache_dir="~/.cache/titanoboa"):
    global _disk_cache
    if cache_dir is None:
        _disk_cache = None
        return
    compiler_version = f"{vyper.__version__}.{vyper.__commit__}"
    _disk_cache = DiskCache(cache_dir, compiler_version)


def disable_cache():
    set_cache_dir(None)


set_cache_dir()  # enable caching, by default!


class BoaImporter(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):
        path = Path(fullname.replace(".", "/")).with_suffix(".vy")

        for prefix in sys.path:
            to_try = Path(prefix) / path

            if to_try.exists():
                loader = BoaLoader(fullname, str(to_try))
                return spec_from_loader(fullname, loader)


class BoaLoader(SourceFileLoader):
    def get_code(self, fullname):
        # importlib docs say to return None, but that triggers an `ImportError`
        return ""

    def create_module(self, spec):
        ret = load_partial(self.path)

        # comply with PEP-302:
        ret.__name__ = spec.name
        ret.__file__ = self.path
        ret.__loader__ = self
        ret.__package__ = spec.name.rpartition(".")[0]
        return ret


sys.meta_path.append(BoaImporter())


def hash_input(compiler_input: CompilerInput) -> str:
    if isinstance(compiler_input, FileInput):
        return compiler_input.sha256sum
    if isinstance(compiler_input, ABIInput):
        return sha256sum(str(compiler_input.abi))
    raise RuntimeError(f"bad compiler input {compiler_input}")


# compute a fingerprint for a module which changes if any of its
# dependencies change
# TODO consider putting this in its own module
def get_module_fingerprint(
    module_t: ModuleT, seen: dict["ImportInfo", str] = None
) -> str:
    seen = seen or {}
    fingerprints = []
    for stmt in module_t.import_stmts:
        import_info = stmt._metadata["import_info"]
        if id(import_info) not in seen:
            if isinstance(import_info.typ, ModuleT):
                fingerprint = get_module_fingerprint(import_info.typ, seen)
            else:
                fingerprint = hash_input(import_info.compiler_input)
            seen[id(import_info)] = fingerprint
        fingerprint = seen[id(import_info)]
        fingerprints.append(fingerprint)
    fingerprints.append(module_t._module.source_sha256sum)

    return sha256sum("".join(fingerprints))


def compiler_data(
    source_code: str, contract_name: str, filename: str | Path, deployer=None, **kwargs
) -> CompilerData:
    global _disk_cache, _search_path

    file_input = FileInput(
        contents=source_code,
        source_id=-1,
        path=Path(contract_name),
        resolved_path=Path(filename),
    )
    search_paths = get_search_paths(_search_path)
    input_bundle = FilesystemInputBundle(search_paths)

    settings = Settings(**kwargs)
    ret = CompilerData(file_input, input_bundle, settings)
    if _disk_cache is None:
        return ret

    with anchor_settings(ret.settings):
        # note that this actually parses and analyzes all dependencies,
        # even if they haven't changed. an optimization would be to
        # somehow convince vyper (in ModuleAnalyzer) to get the module_t
        # from the cache.
        module_t = ret.annotated_vyper_module._metadata["type"]
    fingerprint = get_module_fingerprint(module_t)

    def get_compiler_data():
        with anchor_settings(ret.settings):
            # force compilation to happen so DiskCache will cache the compiled artifact:
            _ = ret.bytecode, ret.bytecode_runtime
        return ret

    assert isinstance(deployer, type)
    deployer_id = repr(deployer)  # a unique str identifying the deployer class
    cache_key = str((contract_name, fingerprint, kwargs, deployer_id))
    return _disk_cache.caching_lookup(cache_key, get_compiler_data)


def load(filename: str | Path, *args, **kwargs) -> _Contract:  # type: ignore
    name = Path(filename).stem
    # TODO: investigate if we can just put name in the signature
    if "name" in kwargs:
        name = kwargs.pop("name")
    with open(filename) as f:
        return loads(f.read(), *args, name=name, **kwargs, filename=filename)


def loads(
    source_code,
    *args,
    as_blueprint=False,
    name=None,
    filename=None,
    compiler_args=None,
    **kwargs,
):
    d = loads_partial(source_code, name, filename=filename, compiler_args=compiler_args)
    if as_blueprint:
        return d.deploy_as_blueprint(**kwargs)
    else:
        return d.deploy(*args, **kwargs)


def load_abi(filename: str, *args, name: str = None, **kwargs) -> ABIContractFactory:
    if name is None:
        name = filename
    with open(filename) as fp:
        return loads_abi(fp.read(), *args, name=name, **kwargs)


def loads_abi(json_str: str, *args, name: str = None, **kwargs) -> ABIContractFactory:
    return ABIContractFactory.from_abi_dict(json.loads(json_str), name, *args, **kwargs)


def loads_partial(
    source_code: str,
    name: str = None,
    filename: str | Path | None = None,
    dedent: bool = True,
    compiler_args: dict = None,
) -> VyperDeployer:
    name = name or "VyperContract"  # TODO handle this upstream in CompilerData
    filename = filename or "<unknown>"
    if dedent:
        source_code = textwrap.dedent(source_code)

    compiler_args = compiler_args or {}

    deployer_class = _get_default_deployer_class()
    data = compiler_data(source_code, name, filename, deployer_class, **compiler_args)
    return deployer_class(data, filename=filename)


def load_partial(filename: str, compiler_args=None):
    with open(filename) as f:
        return loads_partial(
            f.read(), name=filename, filename=filename, compiler_args=compiler_args
        )


def from_etherscan(
    address: Any, name=None, uri="https://api.etherscan.io/api", api_key=None
):
    addr = Address(address)
    abi = fetch_abi_from_etherscan(addr, uri, api_key)
    return ABIContractFactory.from_abi_dict(abi, name=name).at(addr)


def _get_default_deployer_class():
    env = Env.get_singleton()
    if hasattr(env, "deployer_class"):
        return env.deployer_class
    return VyperDeployer


__all__ = []  # type: ignore



import contextlib
import hashlib
import os
import pickle
import threading
import time
from pathlib import Path

_ONE_WEEK = 7 * 24 * 3600


@contextlib.contextmanager
# silence errors which can be thrown when handling a file that does
# not exist
def _silence_io_errors():
    try:
        yield
    # pypy throws FileNotFoundError
    except (OSError, FileNotFoundError):  # noqa: B014
        pass


class DiskCache:
    def __init__(self, cache_dir, version_salt, ttl=_ONE_WEEK):
        self.cache_dir = Path(cache_dir).expanduser()
        self.version_salt = version_salt
        self.ttl = ttl

        self.last_gc = 0

    def gc(self, force=False):
        for root, dirs, files in os.walk(self.cache_dir):
            # delete items older than ttl
            for f in files:
                # squash errors, file might have been removed in race
                # (both p.stat() and p.unlink() can throw)
                with _silence_io_errors():
                    p = Path(root).joinpath(Path(f))
                    if time.time() - p.stat().st_atime > self.ttl or force:
                        p.unlink()

            # prune empty directories
            for d in dirs:
                # squash errors, directory might have been removed in race
                with _silence_io_errors():
                    Path(root).joinpath(Path(d)).rmdir()

        self.last_gc = time.time()

    # content-addressable location
    def cal(self, string):
        preimage = (self.version_salt + string).encode("utf-8")
        digest = hashlib.sha256(preimage).digest().hex()
        return self.cache_dir.joinpath(f"{self.version_salt}/{digest}.pickle")

    # look up x in the cal; on a miss, write back to the cache
    def caching_lookup(self, string, func):
        gc_interval = self.ttl // 10
        if time.time() - self.last_gc >= gc_interval:
            self.gc()

        p = self.cal(string)
        p.parent.mkdir(parents=True, exist_ok=True)
        try:
            with p.open("rb") as f:
                return pickle.loads(f.read())
        except OSError:
            pass  # discard the stack trace in case of other errors

        res = func()
        # use process ID and thread ID to avoid race conditions
        job_id = f"{os.getpid()}.{threading.get_ident()}"
        tmp_p = p.with_suffix(f".{job_id}.unfinished")
        with tmp_p.open("wb") as f:
            f.write(pickle.dumps(res))
        # rename is atomic, don't really need to care about fsync
        # because worst case we will just rebuild the item
        tmp_p.rename(p)
        return res



"""
A wrapper for the py-evm engine to provide a consistent interface for the environment.
Handles low level details around state and py-evm tracing.
"""

import contextlib
import logging
import sys
import warnings
from typing import Any, Iterator, Optional, Type

import eth.constants as constants
import eth.tools.builder.chain as chain
import eth.vm.forks.spurious_dragon.computation as spurious_dragon
from eth._utils.address import generate_contract_address
from eth.chains.mainnet import MainnetChain
from eth.db.account import AccountDB
from eth.db.atomic import AtomicDB
from eth.exceptions import Halt
from eth.vm.code_stream import CodeStream
from eth.vm.gas_meter import allow_negative_refund_strategy
from eth.vm.message import Message
from eth.vm.opcode_values import STOP
from eth.vm.transaction_context import BaseTransactionContext
from eth_utils import setup_DEBUG2_logging

from boa.rpc import RPC
from boa.util.abi import Address, abi_decode
from boa.util.eip1167 import extract_eip1167_address, is_eip1167_contract
from boa.vm.fast_accountdb import patch_pyevm_state_object, unpatch_pyevm_state_object
from boa.vm.fork import AccountDBFork
from boa.vm.gas_meters import GasMeter
from boa.vm.utils import to_bytes, to_int


def enable_pyevm_verbose_logging():
    logging.basicConfig()
    logger = logging.getLogger("eth.vm.computation.Computation")
    setup_DEBUG2_logging()
    logger.setLevel("DEBUG2")


class VMPatcher:
    _exc_patchables = {
        # env vars vyper supports
        "block_number": "_block_number",
        "timestamp": "_timestamp",
        "coinbase": "_coinbase",
        "difficulty": "_difficulty",
        "prev_hashes": "_prev_hashes",
        "chain_id": "_chain_id",
        "gas_limit": "_gas_limit",
        "prevrandao": "_mix_hash",
    }

    _cmp_patchables = {"code_size_limit": "EIP170_CODE_SIZE_LIMIT"}

    def __init__(self, vm):
        patchables = [
            (self._exc_patchables, vm.state.execution_context),
            (self._cmp_patchables, spurious_dragon),
        ]
        # https://stackoverflow.com/a/12999019
        object.__setattr__(self, "_patchables", patchables)

    def __getattr__(self, attr):
        for s, p in self._patchables:
            if attr in s:
                return getattr(p, s[attr])
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        for s, p in self._patchables:
            if attr in s:
                setattr(p, s[attr], value)
                return

    # to help auto-complete
    def __dir__(self):
        patchable_keys = [k for p, _ in self._patchables for k in p]
        return dir(super()) + patchable_keys

    # save and restore patch values
    @contextlib.contextmanager
    def anchor(self):
        snap = {}
        for s, _ in self._patchables:
            for attr in s:
                snap[attr] = getattr(self, attr)

        try:
            yield

        finally:
            for s, _ in self._patchables:
                for attr in s:
                    setattr(self, attr, snap[attr])


_opcode_overrides = {}


def patch_opcode(opcode_value, fn):
    global _opcode_overrides
    _opcode_overrides[opcode_value] = fn


# _precompiles is a global which is loaded to the env computation
# every time one is created. the reasoning is that it would seem
# confusing to have registered precompiles not persist across envs -
# if somebody registers a precompile, presumably they want it to work
# on all envs.
_precompiles = {}


def register_precompile(*args, **kwargs):
    warnings.warn(
        "register_recompile has been renamed to register_raw_precompile!", stacklevel=2
    )


def register_raw_precompile(address, fn, force=False):
    global _precompiles
    address = Address(address)
    if address in _precompiles and not force:
        raise ValueError(f"Already registered: {address}")
    _precompiles[address.canonical_address] = fn


def deregister_raw_precompile(address, force=True):
    address = Address(address).canonical_address
    if address not in _precompiles and not force:
        raise ValueError("Not registered: {address}")
    _precompiles.pop(address, None)


def console_log(computation):
    msgdata = computation.msg.data_as_bytes
    schema, payload = abi_decode("(string,bytes)", msgdata[4:])
    data = abi_decode(schema, payload)
    print(*data, file=sys.stderr)
    return computation


CONSOLE_ADDRESS = bytes.fromhex("000000000000000000636F6E736F6C652E6C6F67")

register_raw_precompile(CONSOLE_ADDRESS, console_log)


# a code stream which keeps a trace of opcodes it has executed
class TracingCodeStream(CodeStream):
    __slots__ = [
        "_length_cache",
        "_fake_codesize",
        "_raw_code_bytes",
        "invalid_positions",
        "valid_positions",
        "program_counter",
    ]

    def __init__(self, *args, start_pc=0, fake_codesize=None, contract=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._trace = []  # trace of opcodes that were run
        self.program_counter = start_pc  # configurable start PC
        self._fake_codesize = fake_codesize  # what CODESIZE returns

    def __iter__(self) -> Iterator[int]:
        # upstream says: "a very performance-sensitive method"
        # note: not clear to me that len(raw_code_bytes) is a hotspot
        while self.program_counter < self._length_cache:
            opcode = self._raw_code_bytes[self.program_counter]

            self._trace.append(self.program_counter)
            self.program_counter += 1
            yield opcode

        yield STOP

    def __len__(self):
        if self._fake_codesize is not None:
            return self._fake_codesize
        return self._length_cache


class Sha3PreimageTracer:
    mnemonic = "SHA3"

    # trace preimages of sha3

    def __init__(self, sha3_op, env):
        self.env = env
        self.sha3 = sha3_op

    def __call__(self, computation):
        size, offset = [to_int(x) for x in computation._stack.values[-2:]]

        # dispatch into py-evm
        self.sha3(computation)

        if size != 64:
            return

        preimage = computation._memory.read_bytes(offset, size)

        value = computation._stack.values[-1]
        image = to_bytes(value)

        self.env.sha3_trace[preimage] = image


class SstoreTracer:
    mnemonic = "SSTORE"

    def __init__(self, sstore_op, env):
        self.env = env
        self.sstore = sstore_op

    def __call__(self, computation):
        value, slot = [to_int(t) for t in computation._stack.values[-2:]]
        account = computation.msg.storage_address

        # we don't want to deal with snapshots/commits/reverts, so just
        # register that the slot was touched and downstream can filter
        # zero entries.
        self.env.sstore_trace.setdefault(account, set()).add(slot)

        # dispatch into py-evm
        self.sstore(computation)


# ### End section: sha3 tracing


# py-evm uses class instantiaters which need to be classes
# instead of like factories or other easier to use architectures -
# `titanoboa_computation` is a class which can be constructed dynamically
class titanoboa_computation:
    _gas_meter_class = GasMeter

    def __init__(self, *args, **kwargs):
        # super() hardcodes CodeStream into the ctor
        # so we have to override it here
        super().__init__(*args, **kwargs)

        self.code = TracingCodeStream(
            self.code._raw_code_bytes,
            fake_codesize=getattr(self.msg, "_fake_codesize", None),
            start_pc=getattr(self.msg, "_start_pc", 0),
        )
        global _precompiles
        # copy so as not to mess with class state
        self._precompiles = self._precompiles.copy()
        self._precompiles.update(_precompiles)

        global _opcode_overrides
        # copy so as not to mess with class state
        self.opcodes = self.opcodes.copy()
        self.opcodes.update(_opcode_overrides)

        self._gas_meter = self._gas_meter_class(
            self.msg.gas, refund_strategy=allow_negative_refund_strategy
        )
        if hasattr(self._gas_meter, "_set_code"):
            self._gas_meter._set_code(self.code)

        self._child_pcs = []
        self._contract_repr_before_revert = None

    @property
    def net_gas_used(self):
        return max(0, self.get_gas_used() - self.get_gas_refund())

    def add_child_computation(self, child_computation):
        super().add_child_computation(child_computation)
        # track PCs of child calls for profiling purposes
        self._child_pcs.append(self.code.program_counter)

    # hijack creations to automatically generate blueprints
    @classmethod
    def apply_create_message(cls, state, msg, tx_ctx, **kwargs):
        computation = super().apply_create_message(state, msg, tx_ctx, **kwargs)

        bytecode = msg.code
        # cf. eth/vm/logic/system/Create* opcodes
        contract_address = msg.storage_address

        if is_eip1167_contract(bytecode):
            contract_address = extract_eip1167_address(bytecode)
            bytecode = cls.env.evm.vm.state.get_code(contract_address)

        if bytecode in cls.env._code_registry:
            target = cls.env._code_registry[bytecode].deployer.at(contract_address)
            target.created_from = Address(msg.sender)
            cls.env.register_contract(contract_address, target)

        return computation

    @classmethod
    def apply_computation(cls, state, msg, tx_ctx, **kwargs):
        addr = msg.code_address
        contract = cls.env._lookup_contract_fast(addr) if addr else None

        def finalize(c):
            if c.is_error:
                # After the computation is applied with an error the state is
                # reverted. Before the revert, save the contract repr for the
                # error message
                c._contract_repr_before_revert = repr(contract)
            return c

        if contract is None or not cls.env.evm._fast_mode_enabled:
            # print("SLOW MODE")
            computation = super().apply_computation(state, msg, tx_ctx, **kwargs)
            return finalize(computation)

        with cls(state, msg, tx_ctx) as computation:
            try:
                if getattr(msg, "_ir_executor", None) is not None:
                    # print("MSG HAS IR EXECUTOR")
                    # this happens when bytecode is overridden, e.g.
                    # for injected functions. note ir_executor is (correctly)
                    # used for the outer computation only because on subcalls
                    # a clean message is constructed for the child computation
                    msg._ir_executor.exec(computation)
                else:
                    # print("REGULAR FAST MODE")
                    contract.ir_executor.exec(computation)
            except Halt:
                pass

        # return computation outside of with block; computation.__exit__
        # swallows exceptions (including Revert).
        return finalize(computation)


# Message object with extra attrs we can use to thread things through
# the execution context.
class FakeMessage(Message):
    def __init__(
        self,
        *args,
        ir_executor=None,
        fake_codesize=None,
        start_pc=0,
        contract=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._ir_executor = ir_executor
        self._fake_codesize = fake_codesize
        self._start_pc = start_pc
        self._contract = contract


class PyEVM:
    def __init__(self, env, fast_mode_enabled=False, fork_try_prefetch_state=False):
        self.chain = _make_chain()
        self.env = env
        self._fast_mode_enabled = fast_mode_enabled
        self._fork_try_prefetch_state = fork_try_prefetch_state
        self._init_vm()

    def _init_vm(self, account_db_class=AccountDB):
        self.vm = self.chain.get_vm()
        self.vm.__class__._state_class.account_db_class = account_db_class

        self.patch = VMPatcher(self.vm)

        c: Type[titanoboa_computation] = type(
            "TitanoboaComputation",
            (titanoboa_computation, self.vm.state.computation_class),
            {"env": self.env},
        )

        if self._fast_mode_enabled:
            patch_pyevm_state_object(self.vm.state)

        self.vm.state.computation_class = c

        # patch in tracing opcodes
        c.opcodes[0x20] = Sha3PreimageTracer(c.opcodes[0x20], self.env)
        c.opcodes[0x55] = SstoreTracer(c.opcodes[0x55], self.env)

    def enable_fast_mode(self, flag: bool = True):
        if flag:
            patch_pyevm_state_object(self.vm.state)
        else:
            unpatch_pyevm_state_object(self.vm.state)

    def fork_rpc(self, rpc: RPC, block_identifier: str, **kwargs):
        account_db_class = AccountDBFork.class_from_rpc(rpc, block_identifier, **kwargs)
        self._init_vm(account_db_class)
        block_info = self.vm.state._account_db._block_info

        self.patch.timestamp = int(block_info["timestamp"], 16)
        self.patch.block_number = int(block_info["number"], 16)
        self.patch.chain_id = int(rpc.fetch("eth_chainId", []), 16)

        self.vm.state._account_db._rpc._init_db()

    @property
    def is_forked(self):
        return issubclass(
            self.vm.__class__._state_class.account_db_class, AccountDBFork
        )

    def get_gas_meter_class(self):
        return self.vm.state.computation_class._gas_meter_class

    def set_gas_meter_class(self, cls: type):
        self.vm.state.computation_class._gas_meter_class = cls

    def get_balance(self, address: Address):
        return self.vm.state.get_balance(address.canonical_address)

    def set_balance(self, address: Address, value):
        self.vm.state.set_balance(address.canonical_address, value)

    def get_code(self, address: Address) -> bytes:
        return self.vm.state.get_code(address.canonical_address)

    def set_code(self, address: Address, code: bytes) -> None:
        self.vm.state.set_code(address.canonical_address, code)

    def get_storage(self, address: Address, slot: int) -> int:
        return self.vm.state.get_storage(address.canonical_address, slot)

    def set_storage(self, address: Address, slot: int, value: int) -> None:
        self.vm.state.set_storage(address.canonical_address, slot, value)

    def get_gas_limit(self):
        return self.vm.state.gas_limit

    # advanced: reset warm/cold counters for addresses and storage
    def reset_access_counters(self):
        self.vm.state._account_db._reset_access_counters()

    def snapshot(self) -> Any:
        return self.vm.state.snapshot()

    def revert(self, snapshot_id: Any) -> None:
        self.vm.state.revert(snapshot_id)

    def generate_create_address(self, sender: Address):
        nonce = self.vm.state.get_nonce(sender.canonical_address)
        self.vm.state.increment_nonce(sender.canonical_address)
        return Address(generate_contract_address(sender.canonical_address, nonce))

    def deploy_code(
        self,
        sender: Address,
        origin: Address,
        target_address: Address,
        gas: Optional[int],
        gas_price: int,
        value: int,
        bytecode: bytes,
    ):
        if gas is None:
            gas = self.vm.state.gas_limit

        msg = Message(
            to=constants.CREATE_CONTRACT_ADDRESS,  # i.e., b""
            sender=sender.canonical_address,
            gas=gas,
            value=value,
            code=bytecode,
            create_address=target_address.canonical_address,
            data=b"",
        )

        if self.is_forked and self._fork_try_prefetch_state:
            self.vm.state._account_db.try_prefetch_state(msg)

        tx_ctx = BaseTransactionContext(
            origin=origin.canonical_address, gas_price=gas_price
        )
        return self.vm.state.computation_class.apply_create_message(
            self.vm.state, msg, tx_ctx
        )

    def execute_code(
        self,
        sender: Address,
        to: Address,
        gas: int,
        gas_price: int,
        value: int,
        bytecode: bytes,
        data: bytes,
        is_static: bool,
        fake_codesize: Optional[int],
        start_pc: int,
        ir_executor: Any,
        contract: Any,
    ):
        msg = FakeMessage(
            sender=sender.canonical_address,
            to=to.canonical_address,
            gas=gas,
            value=value,
            code=bytecode,  # type: ignore
            data=data,
            is_static=is_static,
            fake_codesize=fake_codesize,
            start_pc=start_pc,
            ir_executor=ir_executor,
            contract=contract,
        )

        if self.is_forked and self._fork_try_prefetch_state:
            self.vm.state._account_db.try_prefetch_state(msg)

        origin = sender.canonical_address  # XXX: consider making this parameterizable
        tx_ctx = BaseTransactionContext(origin=origin, gas_price=gas_price)
        return self.vm.state.computation_class.apply_message(self.vm.state, msg, tx_ctx)

    def get_storage_slot(self, address: Address, slot: int) -> bytes:
        data = self.vm.state._account_db.get_storage(address.canonical_address, slot)
        return data.to_bytes(32, "big")


GENESIS_PARAMS = {"difficulty": constants.GENESIS_DIFFICULTY, "gas_limit": int(1e8)}


# TODO make fork configurable - ex. "latest", "frontier", "berlin"
# TODO make genesis params+state configurable
def _make_chain():
    # TODO should we use MiningChain? is there a perf difference?
    # TODO debug why `fork_at()` cannot accept 0 as block num
    _Chain = chain.build(MainnetChain, chain.latest_mainnet_at(1))
    return _Chain.from_genesis(AtomicDB(), GENESIS_PARAMS)



import io
import re
import tokenize
from typing import Any, Optional

import vyper.ast as vy_ast


def get_block(source_code: str, lineno: int, end_lineno: int) -> str:
    source_lines = source_code.splitlines(keepends=True)
    return "".join(source_lines[lineno - 1 : end_lineno])


def get_line(source_code: str, lineno: int) -> str:
    return get_block(source_code, lineno, lineno)


def _get_comment(source_line: str) -> Optional[str]:
    tokens = tokenize.generate_tokens(io.StringIO(source_line).readline)
    return next((t.string for t in tokens if t.type == tokenize.COMMENT), None)


# loosely, match # `@dev asdf...` or `dev: asdf...`
REASON_PATTERN = re.compile(r"#\s*@?(\w+):?\s+(.*)")


def _extract_reason(comment: str) -> Any:
    m = REASON_PATTERN.match(comment)
    if m is not None:
        return m.group(1, 2)
    return None


# extract the dev revert reason at a given line.
# somewhat heuristic.
def reason_at(
    source_code: str, lineno: int, end_lineno: int
) -> Optional[tuple[str, str]]:
    block = get_block(source_code, lineno, end_lineno)
    c = _get_comment(block)
    if c is not None:
        return _extract_reason(c)
    return None


def get_fn_name_from_lineno(ast_map: dict, lineno: int) -> str:
    # TODO: this could be a performance bottleneck
    for source_map, node in ast_map.items():
        if source_map[0] == lineno:
            fn_node = get_fn_ancestor_from_node(node)
            if fn_node:
                return fn_node.name
    return ""


def get_fn_ancestor_from_node(node):
    if node is None:
        return None

    if isinstance(node, vy_ast.FunctionDef):
        return node

    return node.get_ancestor(vy_ast.FunctionDef)



# copy py-evm LevelDB implementation, removed in
# https://github.com/ethereum/py-evm/commit/c2ca44a1212a287
import plyvel
from eth.db.backends.base import BaseDB


class LevelDB(BaseDB):
    _GLOBAL = None

    def __init__(self, db_path, max_open_files: int = None) -> None:
        self.db = plyvel.DB(
            db_path,
            create_if_missing=True,
            error_if_exists=False,
            max_open_files=max_open_files,
        )

    @classmethod
    # Creates db as a class variable to avoid level db lock error
    # create the singleton db object
    def create(cls, *args, **kwargs):
        if cls._GLOBAL is None:
            cls._GLOBAL = cls(*args, **kwargs)
        return cls._GLOBAL

    def __getitem__(self, key: bytes) -> bytes:
        v = self.db.get(key)
        if v is None:
            raise KeyError(key)
        return v

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self.db.put(key, value)

    def _exists(self, key: bytes) -> bool:
        return self.db.get(key) is not None

    def __delitem__(self, key: bytes) -> None:
        if self.db.get(key) is None:
            raise KeyError(key)
        self.db.delete(key)



from datetime import timedelta

NUL = b"\0"
CALLBACK_TOKEN_TIMEOUT = timedelta(minutes=3)
SHARED_MEMORY_LENGTH = 100 * 1024 + len(NUL)  # Size of the shared memory object
CALLBACK_TOKEN_CHARS = 30  # OSx limits this to 31 characters
PLUGIN_NAME = "titanoboa_jupyterlab"
TOKEN_REGEX = rf"[0-9a-fA-F]{{{CALLBACK_TOKEN_CHARS}}}"
TRANSACTION_TIMEOUT_MESSAGE = (
    "Timeout waiting for user to confirm transaction in the browser wallet plug-in."
)
ADDRESS_TIMEOUT_MESSAGE = "Timeout loading browser browser wallet plug-in."
RPC_TIMEOUT_MESSAGE = "Timeout waiting for response from RPC."



import os
import sys
from typing import Any, Type

import rlp
from eth.db.account import AccountDB, keccak
from eth.db.backends.memory import MemoryDB
from eth.db.cache import CacheDB
from eth.db.journal import JournalDB
from eth.rlp.accounts import Account
from eth.vm.interrupt import MissingBytecode
from eth.vm.message import Message
from eth_utils import int_to_big_endian, to_canonical_address, to_checksum_address
from requests import HTTPError

from boa.rpc import RPC, RPCError, fixup_dict, json, to_bytes, to_hex, to_int
from boa.util.lrudict import lrudict

TIMEOUT = 60  # default timeout for http requests in seconds


DEFAULT_CACHE_DIR = "~/.cache/titanoboa/fork.db"
_PREDEFINED_BLOCKS = {"safe", "latest", "finalized", "pending", "earliest"}


_EMPTY = b""  # empty rlp stuff
_HAS_KEY = b"\x01"  # could be anything


class CachingRPC(RPC):
    def __init__(self, rpc: RPC, cache_file: str = DEFAULT_CACHE_DIR):
        # (default to memory db plyvel not found or cache_file is None)
        self._rpc = rpc

        self._cache_file = cache_file
        self._init_db()

    # _loaded is a cache for the constructor.
    # reduces fork time after the first fork.
    _loaded: dict[tuple[str, str], "CachingRPC"] = {}
    _pid: int = os.getpid()  # so we can detect if our fds are bad

    def _init_db(self):
        if self._cache_file is not None:
            try:
                from boa.util.leveldb import LevelDB

                print("(using leveldb)", file=sys.stderr)

                cache_file = os.path.expanduser(self._cache_file)
                # use CacheDB as an additional layer over disk
                # (ideally would use leveldb lru cache but it's not configurable
                # via LevelDB API).
                leveldb = LevelDB.create(cache_file)
                self._db = CacheDB(leveldb, cache_size=1024 * 1024)  # type: ignore
                return
            except ImportError:
                # plyvel not found
                pass

        self._db = MemoryDB(lrudict(1024 * 1024))

    @property
    def identifier(self) -> str:
        return self._rpc.identifier

    @property
    def name(self):
        return self._rpc.name

    def __new__(cls, rpc, cache_file=DEFAULT_CACHE_DIR):
        if isinstance(rpc, cls):
            return rpc

        if os.getpid() != cls._pid:
            # we are in a fork. reload everything so that fds are not corrupted
            cls._loaded = {}
            cls._pid = os.getpid()

        if (rpc.identifier, cache_file) in cls._loaded:
            return cls._loaded[(rpc.identifier, cache_file)]

        ret = super().__new__(cls)
        ret.__init__(rpc, cache_file)
        cls._loaded[(rpc.identifier, cache_file)] = ret
        return ret

    # a stupid key for the kv store
    def _mk_key(self, method: str, params: Any) -> Any:
        return json.dumps({"method": method, "params": params}).encode("utf-8")

    def fetch(self, method, params):
        # cannot dispatch into fetch_multi, doesn't work for debug_traceCall.
        key = self._mk_key(method, params)
        if key in self._db:
            return json.loads(self._db[key])

        result = self._rpc.fetch(method, params)
        self._db[key] = json.dumps(result).encode("utf-8")
        return result

    def fetch_uncached(self, method, params):
        return self._rpc.fetch_uncached(method, params)

    # caching fetch of multiple payloads
    def fetch_multi(self, payload):
        ret = {}
        keys = []
        batch = []
        for item_ix, (method, params) in enumerate(payload):
            key = self._mk_key(method, params)
            try:
                ret[item_ix] = json.loads(self._db[key])
            except KeyError:
                keys.append((key, item_ix))
                batch.append((method, params))

        if len(batch) > 0:
            # fetch_multi is called only with the missing payloads
            # map the results back to the original indices
            for result_ix, rpc_result in enumerate(self._rpc.fetch_multi(batch)):
                key, item_ix = keys[result_ix]
                ret[item_ix] = rpc_result
                self._db[key] = json.dumps(rpc_result).encode("utf-8")

        return [ret[i] for i in range(len(ret))]


# AccountDB which dispatches to an RPC when we don't have the
# data locally
class AccountDBFork(AccountDB):
    @classmethod
    def class_from_rpc(
        cls, rpc: RPC, block_identifier: str, **kwargs
    ) -> Type["AccountDBFork"]:
        class _ConfiguredAccountDB(AccountDBFork):
            def __init__(self, *args, **kwargs2):
                caching_rpc = CachingRPC(rpc, **kwargs)
                super().__init__(caching_rpc, block_identifier, *args, **kwargs2)

        return _ConfiguredAccountDB

    def __init__(self, rpc: CachingRPC, block_identifier: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._dontfetch = JournalDB(MemoryDB())

        self._rpc = rpc

        if block_identifier not in _PREDEFINED_BLOCKS:
            block_identifier = to_hex(block_identifier)

        self._block_info = self._rpc.fetch_uncached(
            "eth_getBlockByNumber", [block_identifier, False]
        )
        self._block_number = to_int(self._block_info["number"])

    @property
    def _block_id(self):
        return to_hex(self._block_number)

    def _has_account(self, address, from_journal=True):
        return super()._get_encoded_account(address, from_journal) != _EMPTY

    def _get_account_helper(self, address, from_journal=True):
        # cf. super impl of _get_account
        if from_journal and address in self._account_cache:
            return self._account_cache[address]

        rlp_account = self._get_encoded_account(address, from_journal)

        if rlp_account:
            return rlp.decode(rlp_account, sedes=Account)
        else:
            return None

    def _get_account(self, address, from_journal=True):
        # we need to override this in order so that internal uses of
        # _set_account() work correctly
        account = self._get_account_helper(address, from_journal)

        if account is None:
            account = self._get_account_rpc(address)

        if from_journal:
            self._account_cache[address] = account

        return account

    def _get_account_rpc(self, address):
        addr = to_checksum_address(address)
        reqs = [
            ("eth_getBalance", [addr, self._block_id]),
            ("eth_getTransactionCount", [addr, self._block_id]),
            ("eth_getCode", [addr, self._block_id]),
        ]
        res = self._rpc.fetch_multi(reqs)
        balance = to_int(res[0])
        nonce = to_int(res[1])
        code = to_bytes(res[2])
        code_hash = keccak(code)

        return Account(nonce=nonce, balance=balance, code_hash=code_hash)

    # try call debug_traceCall to get the ostensible prestate for this call
    def try_prefetch_state(self, msg: Message):
        args = fixup_dict(
            {
                "from": msg.sender,
                "to": msg.to,
                "gas": msg.gas,
                "value": msg.value,
                "data": msg.data,
            }
        )
        try:
            trace_args = [args, self._block_id, {"tracer": "prestateTracer"}]
            trace = self._rpc.fetch("debug_traceCall", trace_args)
        except (RPCError, HTTPError):
            return

        snapshot = self.record()

        for address, account_dict in trace.items():
            try:
                address = to_canonical_address(address)
            except ValueError:
                # the trace we have been given is invalid, roll back changes
                self.discard(snapshot)
                return

            # set account if we don't already have it
            if self._get_account_helper(address) is None:
                balance = to_int(account_dict.get("balance", "0x"))
                code = to_bytes(account_dict.get("code", "0x"))
                nonce = account_dict.get("nonce", 0)  # already an int
                self._set_account(address, Account(nonce=nonce, balance=balance))
                self.set_code(address, code)

            storage = account_dict.get("storage", {})
            for hexslot, hexvalue in storage.items():
                slot = to_int(hexslot)
                value = to_int(hexvalue)
                # set storage if we don't already have it.
                # see AccountStorageDB.get()
                # note we explicitly write 0s, so that they appear
                # in the journal later when called by get_storage
                if not self._helper_have_storage(address, slot):
                    self.set_storage(address, slot, value)

        # the prefetch is lost on later reverts, however the RPC calls are cached
        self.commit(snapshot)

    def get_code(self, address):
        try:
            return super().get_code(address)
        except MissingBytecode:  # will get thrown if code_hash != hash(empty)
            pass

        code_args = [to_checksum_address(address), self._block_id]
        code = to_bytes(self._rpc.fetch("eth_getCode", code_args))
        self.set_code(address, code)
        return code

    def discard(self, checkpoint):
        super().discard(checkpoint)
        self._dontfetch.discard(checkpoint)

    def commit(self, checkpoint):
        super().commit(checkpoint)
        self._dontfetch.commit(checkpoint)

    def record(self):
        checkpoint = super().record()
        self._dontfetch.record(checkpoint)
        return checkpoint

    # helper to determine if something is in the storage db
    # or we need to get from RPC
    def _helper_have_storage(self, address, slot, from_journal=True):
        if not from_journal:
            db = super()._get_address_store(address)._locked_changes
            key = int_to_big_endian(slot)
            return db.get(key, _EMPTY) != _EMPTY

        key = self._get_storage_tracker_key(address, slot)
        return self._dontfetch.get(key) == _HAS_KEY

    def get_storage(self, address, slot, from_journal=True):
        # call super for address warming semantics
        val = super().get_storage(address, slot, from_journal)
        if self._helper_have_storage(address, slot, from_journal=from_journal):
            return val

        fetch_args = [to_checksum_address(address), to_hex(slot), self._block_id]
        val = to_int(self._rpc.fetch("eth_getStorageAt", fetch_args))
        if from_journal:
            # when not from journal, don't override changes
            self.set_storage(address, slot, val)
        return val

    def set_storage(self, address, slot, value):
        super().set_storage(address, slot, value)
        # mark don't fetch
        key = self._get_storage_tracker_key(address, slot)
        self._dontfetch[key] = _HAS_KEY

    def account_exists(self, address):
        if super().account_exists(address):
            return True
        return self.get_balance(address) > 0 or self.get_nonce(address) > 0



import json
import os
from os.path import dirname, join, realpath

from IPython.display import Javascript, display


def install_jupyter_javascript_triggers(debug_mode=False):
    """Run the ethers and titanoboa_jupyterlab Javascript snippets in the browser."""
    cur_dir = dirname(realpath(__file__))
    with open(join(cur_dir, "jupyter.js")) as f:
        js = f.read()

    prefix = os.getenv("JUPYTERHUB_SERVICE_PREFIX", "..")
    js = js.replace("$$JUPYTERHUB_SERVICE_PREFIX", prefix)
    js = js.replace("$$BOA_DEBUG_MODE", json.dumps(debug_mode))

    display(Javascript(js))


def convert_frontend_dict(data):
    """Convert the big integers and filter out empty values."""
    return {
        k: int(v) if isinstance(v, str) and v.isnumeric() else v
        for k, v in data.items()
        if v
    }



import IPython.core.magic as ipython

import boa

_contracts = {}


@ipython.magics_class
class TitanoboaMagic(ipython.Magics):
    @ipython.line_cell_magic
    def vyper(self, line, cell=None):
        if cell is None:
            return boa.eval(line)

        return self.deployer(line, cell)

    # unsure about "vyper" vs "contract" cell magic; keep both until decided
    @ipython.cell_magic
    def deployer(self, line, cell):
        line = line or None
        c = boa.loads_partial(cell, name=line)
        if line:
            self.shell.user_ns[line] = c  # ret available in user ipython locals
            _contracts[line] = c  # ret available in boa.ipython._contracts
        return c

    @ipython.cell_magic
    def contract(self, line, cell):
        line = line or None
        c = boa.loads(cell, name=line)
        if line:
            self.shell.user_ns[line] = c  # ret available in user ipython locals
            _contracts[line] = c  # ret available in boa.ipython._contracts
        return c

    # unsure about "vyper" vs "eval" line magic; keep both until decided
    @ipython.line_magic
    def eval(self, line):
        return boa.eval(line)


def load_ipython_extension(ipy_module):
    ipy_module.register_magics(TitanoboaMagic)



# an Environment which interacts with a real (prod or test) chain
import contextlib
import warnings
from dataclasses import dataclass
from functools import cached_property
from math import ceil

from eth_account import Account
from requests.exceptions import HTTPError

from boa.environment import Env, _AddressType
from boa.rpc import (
    RPC,
    EthereumRPC,
    RPCError,
    fixup_dict,
    to_bytes,
    to_hex,
    to_int,
    trim_dict,
)
from boa.util.abi import Address


class TraceObject:
    def __init__(self, raw_trace):
        self.raw_trace = raw_trace

    @cached_property
    def returndata(self):
        # per debug_traceTransaction, construct a fake computation.
        # hardhat/anvil only give structLogs, alchemy only gives callTracer.
        if "structLogs" in self.raw_trace:
            return self.raw_trace["returnValue"]
        else:
            return self.raw_trace.get("output", "0x")

    @cached_property
    def returndata_bytes(self):
        return to_bytes(self.returndata)

    @cached_property
    def is_error(self):
        if "structLogs" in self.raw_trace:
            return self.raw_trace["failed"]
        # we can have `"error": null` in the payload
        return self.raw_trace.get("error") is not None


class _EstimateGasFailed(Exception):
    pass


@dataclass
class TransactionSettings:
    # when calculating the base fee, the number of blocks N ahead
    # to compute a cap for the Nth block.
    # defaults to 5 (5 blocks ahead, pending block's baseFee * ~1.8)
    # but can be tweaked. if you get errors like
    # `boa.rpc.RPCError: -32000: err: max fee per gas less than block base fee`
    # try increasing the constant.
    # do not recommend setting below 0.
    base_fee_estimator_constant: int = 5

    # amount of time to wait, in seconds before giving up on a transaction
    poll_timeout: float = 240.0


@dataclass
class ExternalAccount:
    address: Address
    _rpc: EthereumRPC

    def __post_init__(self):
        self.address = Address(self.address)

    def send_transaction(self, tx_data):
        txhash = self._rpc.fetch("eth_sendTransaction", [tx_data])
        # format to be the same as what BrowserSigner returns
        return {"hash": txhash}


class Capabilities:
    """
    Describes the capabilities of a chain (right now, EVM opcode support)
    """

    def __init__(self, rpc):
        self._rpc = rpc

    def _get_capability(self, hex_bytecode):
        try:
            self._rpc.fetch("eth_call", [{"to": None, "data": hex_bytecode}])
            return True
        except RPCError:
            return False

    @cached_property
    def has_push0(self):
        # PUSH0
        return self._get_capability("0x5f")

    @cached_property
    def has_mcopy(self):
        # PUSH1 0 DUP1 DUP1 MCOPY
        return self._get_capability("0x600080805E")

    @cached_property
    def has_transient(self):
        # PUSH1 0 DUP1 TLOAD
        return self._get_capability("0x60005C")

    @cached_property
    def has_cancun(self):
        return self.has_shanghai and self.has_mcopy and self.has_transient

    @cached_property
    def has_shanghai(self):
        return self.has_push0

    def describe_capabilities(self):
        if not self.has_shanghai:
            return "pre-shanghai"
        if not self.has_cancun:
            return "shanghai"
        return "cancun"

    def check_evm_version(self, evm_version):
        if evm_version == "cancun":
            return self.has_cancun
        if evm_version == "shanghai":
            return self.has_shanghai
        # don't care about pre-shanghai since there aren't really new
        # opcodes between constantinople and shanghai (and pre-constantinople
        # is no longer even supported by vyper compiler).
        return True


class NetworkEnv(Env):
    """
    An Env object which can be swapped in via `boa.set_env()`.
    It runs non-mutable (view or pure) functions via eth_call,
    mutable functions and contract creation via eth_sendRawTransaction.
    """

    # always prefetch state in network mode
    _fork_try_prefetch_state = True

    def __init__(
        self,
        rpc: str | RPC,
        accounts: dict[str, Account] = None,
        fork_try_prefetch_state=True,
        **kwargs,
    ):
        super().__init__(fork_try_prefetch_state, **kwargs)

        if isinstance(rpc, str):
            warnings.warn(
                "NetworkEnv(url) is deprecated. Use boa.set_network_env(url) instead.",
                stacklevel=2,
            )
            rpc = EthereumRPC(rpc)

        self._rpc: RPC = rpc

        self._reset_fork()

        self._accounts = accounts or {}

        self.eoa = None

        self._gas_price = None

        self.tx_settings = TransactionSettings()
        self.capabilities = Capabilities(rpc)
        self._suppress_debug_tt = False

    @cached_property
    def _rpc_has_snapshot(self):
        try:
            snapshot_id = self._rpc.fetch("evm_snapshot", [])
            self._rpc.fetch("evm_revert", [snapshot_id])
            return True
        except (RPCError, HTTPError):
            return False

    # OVERRIDES
    @contextlib.contextmanager
    def anchor(self):
        if not self._rpc_has_snapshot:
            raise RuntimeError("RPC does not have `evm_snapshot` capability!")
        block_number = self.evm.patch.block_number
        snapshot_id = self._rpc.fetch("evm_snapshot", [])
        try:
            yield
            # note we cannot call super.anchor() because vm/accountdb fork
            # state is reset after every txn.
        finally:
            self._rpc.fetch("evm_revert", [snapshot_id])
            # wipe forked state
            self._reset_fork(block_number)

    # add account, or "Account-like" object. MUST expose
    # `sign_transaction` or `send_transaction` method!
    def add_account(self, account: Account, force_eoa=False):
        self._accounts[account.address] = account  # type: ignore
        if self.eoa is None or force_eoa:
            self.eoa = account.address  # type: ignore

    def add_accounts_from_rpc(self, rpc: str | EthereumRPC) -> None:
        if isinstance(rpc, str):
            rpc = EthereumRPC(rpc)

        # address strings, ex. ["0x0e5437b1b3448d22c07caed31e5bcdc4ec5284a9"]
        addresses = rpc.fetch("eth_accounts", [])
        if not addresses:
            # strip out content in the URL which might not want to get into logs
            warnings.warn(f"No accounts fetched from <{rpc.name}>!", stacklevel=2)
        for address in addresses:
            self.add_account(ExternalAccount(_rpc=rpc, address=address))  # type: ignore

    def set_eoa(self, eoa: Account) -> None:
        self.add_account(eoa, force_eoa=True)

    @classmethod
    def from_url(cls, url: str) -> "NetworkEnv":
        return cls(EthereumRPC(url))

    # overrides
    def get_gas_price(self) -> int:
        if self._gas_price is not None:
            return self._gas_price
        return to_int(self._rpc.fetch("eth_gasPrice", []))

    def get_eip1559_fee(self) -> tuple[str, str, str, str]:
        # returns: base_fee, max_fee, max_priority_fee
        reqs = [
            ("eth_getBlockByNumber", ["latest", False]),
            ("eth_maxPriorityFeePerGas", []),
            ("eth_chainId", []),
        ]
        block_info, max_priority_fee, chain_id = self._rpc.fetch_multi(reqs)
        base_fee = block_info["baseFeePerGas"]

        # Each block increases the base fee by 1/8 at most.
        # here we have the next block's base fee, compute a cap for the
        # next N blocks here.
        blocks_ahead = self.tx_settings.base_fee_estimator_constant
        base_fee_estimate = ceil(to_int(base_fee) * (9 / 8) ** blocks_ahead)

        max_fee = to_hex(base_fee_estimate + to_int(max_priority_fee))
        return to_hex(base_fee_estimate), max_priority_fee, max_fee, chain_id

    def get_static_fee(self) -> tuple[str, str]:
        # non eip-1559 transaction
        return tuple(self._rpc.fetch_multi([("eth_gasPrice", []), ("eth_chainId", [])]))

    def _check_sender(self, address: Address):
        if address is None:
            raise ValueError("No sender!")
        return address

    # OVERRIDES
    def execute_code(
        self,
        to_address,
        sender=None,
        gas=None,
        value=0,
        data=b"",
        override_bytecode=None,
        contract=None,
        is_modifying=True,
        ir_executor=None,  # maybe just have **kwargs to collect extra kwargs
    ):
        if is_modifying:
            # reset to latest block for code simulation
            self._reset_fork()

        # call execute_code for tracing side effects
        # note: we could get a perf improvement if we ran this in
        # the background while waiting on RPC network calls
        computation = super().execute_code(
            to_address=to_address,
            sender=sender,
            gas=gas,
            value=value,
            data=data,
            is_modifying=is_modifying,
            contract=contract,
        )

        sender = self._check_sender(self._get_sender(sender))

        hexdata = to_hex(data)

        if is_modifying:
            try:
                receipt, trace = self._send_txn(
                    from_=sender, to=to_address, value=value, gas=gas, data=hexdata
                )
            except _EstimateGasFailed:
                # no need to actually run the txn.
                # caller will decide what to do with the error - probably revert

                # if not computation.is_error, either a bug in boa
                # or out of sync with node.
                assert computation.is_error
                return computation

            output = None
            if trace is not None:
                output = trace.returndata_bytes
                # gas_used = to_int(receipt["gasUsed"])

                # the node reverted but we didn't. consider this an
                # unrecoverable error and bail out
                if trace.is_error and not computation.is_error:
                    raise RuntimeError(
                        f"panic: local computation succeeded but node didnt: {trace}"
                    )

        else:
            args = fixup_dict(
                {
                    "from": sender,
                    "to": to_address,
                    "gas": gas,
                    "value": value,
                    "data": hexdata,
                }
            )
            returnvalue = self._rpc.fetch("eth_call", [args, "latest"])
            output = to_bytes(returnvalue)
            # we don't need to do the check for computation.is_error
            # because if the eth_call failed it would have just raised
            # an actual RPC error

        # returndata not the same. this means either a bug in
        # titanoboa/py-evm or more likely, state got out of sync.
        # not the greatest, but we will just patch the returndata and
        # pretend nothing happened (which is not really a problem unless
        # the caller wants to inspect the trace or memory).
        if output is not None and computation.output != output:
            warnings.warn(
                "local fork did not match node! this indicates state got out "
                "of sync with the network or a bug in titanoboa!",
                stacklevel=2,
            )
            # just return whatever the node had.
            computation.output = output

        return computation

    # OVERRIDES
    def deploy(self, sender=None, gas=None, value=0, bytecode=b"", **kwargs):
        # reset to latest block for simulation
        self._reset_fork()

        # simulate the deployment
        local_address, computation = super().deploy(
            sender=sender, gas=gas, value=value, bytecode=bytecode
        )
        if computation.is_error:
            return local_address, computation

        if trim_dict(kwargs):
            raise TypeError(f"invalid kwargs to execute_code: {kwargs}")
        bytecode = to_hex(bytecode)
        sender = self._check_sender(self._get_sender(sender))

        receipt, trace = self._send_txn(
            from_=sender, value=value, gas=gas, data=bytecode
        )

        create_address = Address(receipt["contractAddress"])

        if trace is not None and computation.output != trace.returndata_bytes:
            # not sure what to do about this, for now just complain
            warnings.warn(
                "local fork did not match node! this indicates state got out "
                "of sync with the network or a bug in titanoboa!",
                stacklevel=2,
            )
            # return what the node returned anyway
            computation.output = trace.returndata_bytes

        if local_address != create_address:
            raise RuntimeError(f"uh oh! {local_address} != {create_address}")

        # TODO get contract info in here
        print(f"contract deployed at {create_address}")

        return create_address, computation

    @cached_property
    def _tracer(self):
        def _warn_no_tracer():
            warnings.warn(
                "debug_traceTransaction not available! "
                "titanoboa will try hard to interact with the network, but "
                "this means that titanoboa is not able to do certain "
                "safety checks at runtime. it is recommended to switch "
                "to a node or provider with debug_traceTransaction.",
                stacklevel=3,
            )

        try:
            txn_hash = "0x" + "00" * 32
            # alchemy only can do callTracer, plus it has lowest
            # overhead.
            call_tracer = {"tracer": "callTracer", "onlyTopCall": True}
            self._rpc.fetch("debug_traceTransaction", [txn_hash, call_tracer])

        except RPCError as e:
            # can't handle callTracer, use default (i.e. structLogs)
            if e.code == -32602:
                return {}

            # catchall - just don't have a tracer
            # note on error codes:
            # -32600 is alchemy unpaid tier error message
            # -32601 is infura error message (if i recall correctly)
            _warn_no_tracer()
            return None

        except HTTPError:
            _warn_no_tracer()
            return None

        return call_tracer

    def suppress_debug_tt(self, new_value=True):
        self._suppress_debug_tt = new_value

    def _debug_tt(self, tx_hash):
        if self._tracer is None:
            return None
        try:
            return self._rpc.fetch_uncached(
                "debug_traceTransaction", [tx_hash, self._tracer]
            )
        except (HTTPError, RPCError) as e:
            if self._suppress_debug_tt:
                warnings.warn(f"Couldn't get a trace for {tx_hash}!", stacklevel=3)
            else:
                warnings.warn(
                    f"Couldn't get a trace for {tx_hash}! If you want to"
                    " suppress this error, call `boa.env.suppress_debug_tt()`"
                    " first.",
                    stacklevel=3,
                )
                raise e

        return None

    def _get_nonce(self, addr):
        return self._rpc.fetch("eth_getTransactionCount", [addr, "latest"])

    def _reset_fork(self, block_identifier="latest"):
        # use "latest" to make sure we are forking with up-to-date state
        # but use reset_traces=False to help with storage dumps
        self.fork_rpc(
            self._rpc,
            reset_traces=False,
            block_identifier=block_identifier,
            cache_file=None,
        )

    def _send_txn(self, from_, to=None, gas=None, value=None, data=None):
        tx_data = fixup_dict(
            {"from": from_, "to": to, "gas": gas, "value": value, "data": data}
        )

        try:
            # eip-1559 txn
            (base_fee, max_priority_fee, max_fee, chain_id) = self.get_eip1559_fee()
            tx_data["maxPriorityFeePerGas"] = max_priority_fee
            tx_data["maxFeePerGas"] = max_fee
            tx_data["chainId"] = chain_id
        except (RPCError, KeyError) as e:
            warnings.warn(
                "No EIP-1559 transaction available, falling back to legacy",
                stacklevel=3,
            )
            warnings.warn(str(e), stacklevel=3)
            gas_price, chain_id = self.get_static_fee()
            tx_data["gasPrice"] = gas_price
            tx_data["chainId"] = chain_id

        tx_data["nonce"] = self._get_nonce(from_)

        if gas is None:
            try:
                tx_data["gas"] = self._rpc.fetch("eth_estimateGas", [tx_data])
            except RPCError as e:
                if e.code == 3:
                    # execution failed at estimateGas, probably the txn reverted
                    raise _EstimateGasFailed() from e
                raise e from e

        if from_ not in self._accounts:
            raise ValueError(f"Account not available: {from_}")
        account = self._accounts[from_]

        if hasattr(account, "sign_transaction"):
            signed = account.sign_transaction(tx_data)

            # note: signed.rawTransaction has type HexBytes
            tx_hash = self._rpc.fetch(
                "eth_sendRawTransaction", [to_hex(bytes(signed.raw_transaction))]
            )
        else:
            # some providers (i.e. metamask) don't have sign_transaction
            # we just have to call send_transaction and pray for the best
            tx_hash = account.send_transaction(tx_data)["hash"]

        # TODO real logging
        print(f"tx broadcasted: {tx_hash}")

        receipt = self._rpc.wait_for_tx_receipt(tx_hash, self.tx_settings.poll_timeout)
        if receipt.get("status") != "0x1":
            raise Exception(f"txn failed: {receipt}")

        trace = self._debug_tt(tx_hash)

        print(f"{tx_hash} mined in block {receipt['blockHash']}!")

        # the block was mined, reset state
        self._reset_fork(block_identifier=receipt["blockNumber"])

        t_obj = TraceObject(trace) if trace is not None else None
        return receipt, t_obj

    def get_chain_id(self) -> int:
        """Get the current chain ID of the network as an integer."""
        chain_id = self._rpc.fetch("eth_chainId", [])
        return int(chain_id, 16)

    def set_balance(self, address, value):
        raise NotImplementedError("Cannot use set_balance in network mode")

    def set_code(self, address: _AddressType, code: bytes) -> None:
        raise NotImplementedError("Cannot use set_code in network mode")

    def set_storage(self, address: _AddressType, slot: int, value: int) -> None:
        raise NotImplementedError("Cannot use set_storage in network mode")
